// AttributeSet.cpp: implementation of the CAttributeSet class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Normalization.h"
#include "AttributeSet.h"
#include "FDSet.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CAttributeSet::CAttributeSet()
{
	m_rPosition=NULL;
	m_nIndex=0;
}

CAttributeSet::CAttributeSet(CAttribute &attInput)
{
	m_attList.AddTail(attInput);
	m_nIndex=0;
	m_rPosition=m_attList.GetHeadPosition();
}

CAttributeSet::CAttributeSet(CAttributeSet &attsetInput)
{
	*this=attsetInput;
}

CAttributeSet::~CAttributeSet()
{

}
 
int CAttributeSet::GetCount() const
{
	return m_attList.GetCount();
}

void CAttributeSet::GoHeadPosition()
{
	m_rPosition=m_attList.GetHeadPosition();
	m_nIndex=0;
}

void CAttributeSet::GoIndex(int nInput)
{
	if(nInput<0 || nInput>=m_attList.GetCount())
		return;

	m_nIndex=nInput;
	m_rPosition=m_attList.FindIndex(m_nIndex);
}

void CAttributeSet::AddAtt(CAttribute& attInput)
{
	if(IsMemberOf(attInput)==TRUE)
		return;

	m_attList.AddTail(attInput);
}

void CAttributeSet::DelAtt(CAttribute& attInput)
{
	const int nSizeAttSet=m_attList.GetCount();
	CAttribute attTemp;

	this->GoHeadPosition();
	for(int i=0; i<nSizeAttSet; i++)
	{
		POSITION rPosition=m_rPosition;

		attTemp=this->GetAtt();
		this->operator++();

		if(attInput==attTemp)
		{
			m_attList.RemoveAt(rPosition);
			break;
		}
	}
}

CAttribute CAttributeSet::GetAtt() const
{
	return m_attList.GetAt(m_rPosition);
}

BOOL CAttributeSet::IsMemberOf(CAttribute attInput)
{
	return IsMemberOf((CAttributeSet)attInput);
}

BOOL CAttributeSet::IsMemberOf(CAttributeSet attsetInput)
{
	if(m_attList.GetCount()<=0)
		return FALSE;
	if(m_attList.GetCount()<attsetInput.GetCount())
		return FALSE;

	BOOL bTrue;

	CAttribute attTemp1;
	CAttribute attTemp2;
	
	attsetInput.GoHeadPosition();
	for(int i=0; i<attsetInput.GetCount(); i++)
	{
		attTemp1=attsetInput.GetAtt();
		attsetInput++;

		GoHeadPosition();
		for(int j=0; j<m_attList.GetCount(); j++)
		{
			attTemp2=GetAtt();
			operator++();

			if(attTemp1!=attTemp2)
				bTrue=FALSE;
			else
			{
				bTrue=TRUE;
				break;
			}
		}
		if(bTrue==FALSE)
			return FALSE;
	}

	return TRUE;
}

CAttributeSet CAttributeSet::Closure(CFDSet fdsetInput)
{
	int nLengthList=fdsetInput.GetCount();
	CAttributeSet attsetClosure=*this;
	BOOL bChange;
	CFD fdTemp1;

	while(1)
	{
		bChange=FALSE;

		fdsetInput.GoHeadPosition();
		for(int i=0; i<nLengthList; i++)
		{
			fdTemp1=fdsetInput.GetFD();
			fdsetInput++;
			if(attsetClosure.IsMemberOf(fdTemp1.GetLHS())==TRUE)
			{
				if(attsetClosure.IsMemberOf(fdTemp1.GetRHS())==FALSE)
				{
					attsetClosure+=fdTemp1.GetRHS();
					bChange=TRUE;
				}
			}
		}
		if(bChange==FALSE)
			break;
	}
	return attsetClosure;
}

CAttributeSet CAttributeSet::GetSubSet(int i)
{
	const int nBackupIndex=m_nIndex;
	const POSITION rBackupPosition=m_rPosition;
	const nSizeAttSet=m_attList.GetCount();

	CAttribute attX;
	CAttributeSet attsetG;

	int* code=new int[nSizeAttSet];

	for(int j=0; j!=nSizeAttSet; j++)
	{
		if((i%2)==1)
			code[j]=1;
		else
			if((i%2)==0)
				code[j]=0;

		i=i/2;
	}

	int nElemnum=0;

	this->GoHeadPosition();
	for(j=0; j<nSizeAttSet; j++)
	{
		attX=this->GetAtt();
		this->operator++();

		if(code[nElemnum]==0);
		else
			if(code[nElemnum]==1)
				attsetG.AddAtt(attX);
		if(nElemnum!=nSizeAttSet-1)
			nElemnum++;
	}

	delete[] code;

	m_nIndex=nBackupIndex;
	m_rPosition=rBackupPosition;
	return attsetG;
}

CAttributeSet CAttributeSet::GetSubSet(int i, int size)
{
	const int nBackupIndex=m_nIndex;
	const POSITION rBackupPosition=m_rPosition;

	CAttribute attX;
	CAttributeSet attsetG;

	int* code=new int[size];

	for(int j=0; j!=size; j++)
	{
		if((i%2)==1)
			code[j]=1;
		else
			if((i%2)==0)
				code[j]=0;

		i=i/2;
	}

//	int nElemnum=size-1;
	int nElemnum=0;
	const int nSizeFDSet=m_attList.GetCount();
	this->GoHeadPosition();
	for(j=0; j<nSizeFDSet; j++)
	{
		attX=this->GetAtt();
		this->operator++();

		if(code[nElemnum]==0);
		else
			if(code[nElemnum]==1)
				attsetG.AddAtt(attX);
		if(nElemnum!=size-1)
			nElemnum++;
//		if(nElemnum!=0)
//			nElemnum--;
	}

	delete[] code;

	m_nIndex=nBackupIndex;
	m_rPosition=rBackupPosition;
	return attsetG;
}

void CAttributeSet::RemoveAtt(int nIndex)
{
	POSITION rPosition;
	rPosition=m_attList.FindIndex(nIndex);
	m_attList.RemoveAt(rPosition);
}

POSITION CAttributeSet::Find(CAttribute attInput)
{
	return m_attList.Find(attInput);
}

BOOL CAttributeSet::IsEmpty() const
{
	return m_attList.IsEmpty();
}

void CAttributeSet::Empty()
{
	m_attList.RemoveAll();
	m_nIndex=0;
	m_rPosition=m_attList.GetHeadPosition();
}

CAttributeSet CAttributeSet::operator+(CAttributeSet& attsetInput)
{
	CAttribute attTemp;
	CAttributeSet attsetTemp;

	attsetTemp=*this;

	attsetInput.GoHeadPosition();
	for(int i=0; i<attsetInput.GetCount(); i++)
	{
		attTemp=attsetInput.GetAtt();
		attsetTemp.AddAtt(attTemp);
		attsetInput++;
	}
	
	return attsetTemp;
}

CAttributeSet CAttributeSet::operator-(CAttribute& attInput)
{
	CAttributeSet attsetTemp=*this;
	attsetTemp=attsetTemp-(CAttribute)attInput;
	return attsetTemp;
}

CAttributeSet CAttributeSet::operator-(CAttributeSet& attsetInput)
{
	CAttribute attTemp;
	CAttributeSet attsetTemp=*this;
	attsetInput.GoHeadPosition();
	for(int i=0; i<attsetInput.GetCount(); i++)
	{
		attTemp=attsetInput.GetAtt();
		attsetInput++;
		attsetTemp.DelAtt(attTemp);
	}
	return attsetTemp;
}

const CAttributeSet& CAttributeSet::operator=(CAttribute& attInput)
{
	*this=CAttributeSet(attInput);
	return *this;
}

const CAttributeSet& CAttributeSet::operator=(CAttributeSet& attsetInput)
{
	if(this==&attsetInput)
		return *this;

	CAttribute attTemp;
	m_attList.RemoveAll();

	attsetInput.GoHeadPosition();
	for(int i=0; i<attsetInput.GetCount(); i++)
	{
		attTemp=attsetInput.GetAtt();
		m_attList.AddTail(attTemp);
		attsetInput++;
	}

	m_nIndex=0;
	m_rPosition=m_attList.GetHeadPosition();
	return *this;
}

const CAttributeSet& CAttributeSet::operator +=(CAttribute& attInput)
{
	POSITION rPosition;
	rPosition=m_attList.Find(attInput);
	if(rPosition!=NULL)
		return *this;

	*this=*this+(CAttributeSet)attInput;
	return *this;
}

const CAttributeSet& CAttributeSet::operator +=(CAttributeSet& attsetInput)
{
	*this=*this+attsetInput;
	return *this;
}

void CAttributeSet::operator++()
{
	m_nIndex++;
	if(m_nIndex>=m_attList.GetCount())
		m_nIndex=m_attList.GetCount()-1;
	else
		m_rPosition=m_attList.FindIndex(m_nIndex);
}

void CAttributeSet::operator++(int)
{
	operator++();
}

void CAttributeSet::operator--()
{
	m_nIndex--;
	if(m_nIndex<0)
		m_nIndex=0;
	else
		m_rPosition=m_attList.FindIndex(m_nIndex);
}

void CAttributeSet::operator--(int)
{
	operator--();
}

BOOL CAttributeSet::operator==(CAttributeSet& attsetInput)
{
	if(m_attList.GetCount()!=attsetInput.GetCount())
		return FALSE;

	return IsMemberOf(attsetInput);
}

BOOL CAttributeSet::operator!=(CAttributeSet& attsetInput)
{
	return (!operator==(attsetInput));
}

CString CAttributeSet::GetstrAtt()
{
	CString strTemp;
	const int nListLength=m_attList.GetCount();

	GoHeadPosition();
	for(int i=0; i<nListLength; i++)
	{
		strTemp+=GetAtt().GetAttName()+", ";
		operator++();
	}
	strTemp.Delete(strTemp.GetLength()-2, 2);
//	strTemp+="\r\n";
	return strTemp;
}