// FDSet.cpp: implementation of the CFDSet class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Normalization.h"
#include "FDSet.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CFDSet::CFDSet()
{
	m_nIndex=0;
	m_rPosition=m_fdList.GetHeadPosition();
}

CFDSet::CFDSet(CFD& fdInput)
{
	m_fdList.AddTail(fdInput);
	m_nIndex=0;
	m_rPosition=m_fdList.GetHeadPosition();
}

CFDSet::CFDSet(CFDSet& fdsetInput)
{
	const int nBackupIndex=fdsetInput.GetCount();

	*this=fdsetInput;
	
	fdsetInput.GoIndex(nBackupIndex);
}

CFDSet::~CFDSet()
{

}
  
int CFDSet::GetCount()
{
	return m_fdList.GetCount();
}

void CFDSet::GoHeadPosition()
{
	m_rPosition=m_fdList.GetHeadPosition();
	m_nIndex=0;
}

void CFDSet::GoIndex(int nInput)
{
	if(nInput<0 || nInput>=m_fdList.GetCount())
		return;

	m_nIndex=nInput;
	m_rPosition=m_fdList.FindIndex(m_nIndex);
}

void CFDSet::AddFD(CFD fdInput, int temp=FALSE)
{
	if(temp==FALSE)
		if(IsMemberOf(fdInput)==TRUE)
			return;

	m_fdList.AddTail(fdInput);
}

CFD CFDSet::GetFD()
{
	CFD temp=m_fdList.GetAt(m_rPosition);
	return temp;
}

int CFDSet::GetIndex()
{
	return m_nIndex;
}

// fdset coverϴ Attribute Set ã´.
CAttributeSet CFDSet::GetCover()
{
	CAttributeSet attsetCover;
	const int nSizeFDSet=m_fdList.GetCount();

	if(nSizeFDSet<=0)
		return attsetCover;

	this->GoHeadPosition();
	for(int cI=0; cI<nSizeFDSet; cI++)
	{
		CFD fdX=this->GetFD();
		this->operator ++();

		attsetCover+=fdX.GetLHS()+fdX.GetRHS();
	}

	return attsetCover;
}


// fdset determinant set ã´.
CAttributeSet CFDSet::GetDeterminant()
{
	CAttributeSet attsetDeterminant;
	const int nSizeFDSet=m_fdList.GetCount();

	if(nSizeFDSet<=0)
		return attsetDeterminant;

	this->GoHeadPosition();
	for(int cI=0; cI<nSizeFDSet; cI++)
	{
		CFD fdX=this->GetFD();
		this->operator ++();

		attsetDeterminant+=fdX.GetLHS();
	}

	return attsetDeterminant;
}

void CFDSet::Remove(CFD fdInput)
{
	int nIndex=m_nIndex;
	POSITION rPosition=m_rPosition;

	int nLengthList=m_fdList.GetCount();
	CFD fdTemp;

	GoHeadPosition();
	for(int i=0; i<nLengthList; i++)
	{
		POSITION rPosition=m_rPosition;

		fdTemp=GetFD();
		operator++();

		if(fdTemp==fdInput)
		{
			m_fdList.RemoveAt(rPosition);
			break;
		}
	}

	m_nIndex=nIndex;
	if(m_nIndex>=m_fdList.GetCount())
		m_nIndex=m_fdList.GetCount()-1;
	else
		if(m_nIndex<0)
			m_nIndex=0;
	m_rPosition=m_fdList.FindIndex(m_nIndex);
}

BOOL CFDSet::IsMemberOf(CFD fdInput)
{
	return IsMemberOf((CFDSet)fdInput);
}

BOOL CFDSet::IsMemberOf(CFDSet fdsetInput)
{
	const int nBackupIndex=m_nIndex;
	const POSITION rBackupPosition=m_rPosition;

	if(m_fdList.GetCount()<=0)
		return FALSE;
	if(m_fdList.GetCount()<fdsetInput.GetCount())
		return FALSE;

	BOOL bTrue;

	CFD fdTemp1;
	CFD fdTemp2;

	fdsetInput.GoHeadPosition();
	for(int i=0; i<fdsetInput.GetCount(); i++)
	{
		fdTemp1=fdsetInput.GetFD();
		fdsetInput++;

		GoHeadPosition();
		for(int j=0; j<m_fdList.GetCount(); j++)
		{
			fdTemp2=GetFD();
			operator++();
			if(fdTemp1!=fdTemp2)
				bTrue=FALSE;
			else
			{
				bTrue=TRUE;
				break;
			}
		}
		if(bTrue==FALSE)
		{
			m_nIndex=nBackupIndex;
			m_rPosition=rBackupPosition;
			return FALSE;
		}
	}

	m_nIndex=nBackupIndex;
	m_rPosition=rBackupPosition;
	return TRUE;
}

BOOL CFDSet::IsEmpty()
{
	if(m_fdList.GetCount()==0)
		return TRUE;
	return FALSE;
}

void CFDSet::Empty()
{
	m_fdList.RemoveAll();
}

CFDSet CFDSet::GetSubSet(int i, int size)
{
	const int nBackupIndex=m_nIndex;
	const POSITION rBackupPosition=m_rPosition;
	
	CFD fdX;
	CFDSet fdsetG;

	int* code=new int[size];

	for(int j=0; j!=size; j++){
		if((i%2)==1)
			code[j]=1;
		else
			if((i%2)==0)
				code[j]=0;
		
		i=i/2;
	}

//	int nElemnum=size-1;
	int nElemnum=0;
	const int nLenOfFDList=m_fdList.GetCount();
	this->GoHeadPosition();
	for(j=0; j<nLenOfFDList; j++)
	{
		fdX=this->GetFD();
		this->operator ++();

		if(code[nElemnum]==0);
		else
			if(code[nElemnum]==1)
				fdsetG.AddFD(fdX);

		if(nElemnum!=size-1)
			nElemnum++;
//		if(nElemnum!=0)
//			nElemnum--;
	}

	m_nIndex=nBackupIndex;
	m_rPosition=rBackupPosition;
	return fdsetG;
}


// X-Y X==Yΰ ְ Y=Y-X Ѵ.
//  X=Y=0 ͵ .
CFDSet CFDSet::DelRedundant()
{
	const int nSizeFDSet=m_fdList.GetCount();
	CAttributeSet attsetX;
	CAttributeSet attsetY;
	CFD fdTemp;
	CFDSet fdsetTemp;

	this->GoHeadPosition();
	for(int i=0; i<nSizeFDSet; i++)
	{
		fdTemp=this->GetFD();
		this->operator++();

		attsetX=fdTemp.GetLHS();
		attsetY=fdTemp.GetRHS();
		if(attsetX!=attsetY)
		{
			attsetY=attsetY-attsetX;
			if(attsetY.IsEmpty()==FALSE)
			{
				fdTemp.SetRHS(attsetY);
				fdsetTemp+=fdTemp;
			}
		}
	}

	return fdsetTemp;
}

//2NF鶧 Ѵ.
CFDSet CFDSet::DelRedundant(CAttributeSet attsetInput)
{
	CFD fdTemp;
	CFDSet fdsetTemp;
	CAttributeSet attsetX;

	int nSizeFDSet=m_fdList.GetCount();
	this->GoHeadPosition();
	for(int i=0; i<nSizeFDSet; i++)
	{
		fdTemp=this->GetFD();
		this->operator++();

		attsetX=fdTemp.GetLHS();

		if(attsetInput==attsetX)
		{
			fdsetTemp+=fdTemp;
		}
	}

	fdsetTemp.DelRedundant();
	
	return fdsetTemp;
}

 
CFDSet CFDSet::DelUesless(CAttributeSet attsetInput)
{
	CString strTemp;

	strTemp=attsetInput.GetstrAtt();

	CFDSet fdsetTemp(*this);
	CFDSet fdsetTemp2;

	int nSizeFDSet=fdsetTemp.GetCount();
	fdsetTemp.GoHeadPosition();
	for(int i=0; i<nSizeFDSet; i++)
	{
		CFD fdTemp=fdsetTemp.GetFD();
		fdsetTemp++;

		strTemp=fdTemp.GetstrFD();

		CAttributeSet attsetRHS=fdTemp.GetRHS();
		CAttributeSet attsetLHS=fdTemp.GetLHS();

		strTemp=attsetRHS.GetstrAtt();
		strTemp=attsetLHS.GetstrAtt();

		if(attsetInput.IsMemberOf(attsetRHS)==FALSE)
		{
			CAttributeSet attsetTemp=attsetRHS-attsetInput;

			strTemp=attsetTemp.GetstrAtt();

			attsetTemp=attsetRHS-attsetTemp;

			if(attsetTemp.IsEmpty()==FALSE)
			{
				strTemp=attsetTemp.GetstrAtt();

				fdTemp.SetRHS(attsetTemp);

				strTemp=fdTemp.GetstrFD();

				if(attsetInput.IsMemberOf(attsetLHS)==FALSE)
				{
					continue;
				}

				if(fdTemp.GetLHS().IsMemberOf(fdTemp.GetRHS())==FALSE)
					fdsetTemp2+=fdTemp;
			}
		}
		else
		{
			if(attsetInput.IsMemberOf(attsetLHS)!=FALSE)
			{
			fdsetTemp2+=fdTemp;

			}

		}
	}

	strTemp=fdsetTemp2.GetstrFD();

	return fdsetTemp2;
}

CFDSet CFDSet::NonRedundantCover()
{
	const int nBackupIndex=m_nIndex;
	const POSITION rBackupPosition=m_rPosition;
	
	CFDSet fdsetTemp1=*this;
	CFDSet fdsetTemp2=*this;
	CFD fdTemp1;
	const int nLengthList=fdsetTemp1.GetCount();

	fdsetTemp1.GoHeadPosition();
	for(int i=0; i<nLengthList; i++)
	{
		fdTemp1=fdsetTemp1.GetFD();
		fdsetTemp1++;

		fdsetTemp2.Remove(fdTemp1);
		if(fdTemp1.IsMemberOf(fdsetTemp2))
			fdsetTemp1.Remove(fdTemp1);
	}

	m_nIndex=nBackupIndex;
	m_rPosition=rBackupPosition;
	return fdsetTemp1;
}

CFDSet CFDSet::LeftReducedCover()
{
	const int nBackupIndex=m_nIndex;
	const POSITION rBackupPosition=m_rPosition;

	CFDSet fdsetTemp1=*this;
	CFD fdTemp1;
	CFD fdTemp2;
	CAttribute attTemp;
	CAttributeSet attsetTemp1;
	CAttributeSet attsetTemp2;
	int nLenOfFDList=m_fdList.GetCount();
	int nLenOfAttList;

	GoHeadPosition();
	for(int i=0; i<nLenOfFDList; i++)
	{
		fdTemp1=GetFD();
		operator++();

		attsetTemp1=fdTemp1.GetLHS();

		nLenOfAttList=attsetTemp1.GetCount();
		attsetTemp1.GoHeadPosition();
		for(int j=0; j<nLenOfAttList; j++)
		{
			attTemp=attsetTemp1.GetAtt();
			attsetTemp1++;

			attsetTemp2=attsetTemp1;
			attsetTemp2.DelAtt(attTemp);
			fdTemp2.SetLHS(attsetTemp2);
			fdTemp2.SetRHS(fdTemp1.GetRHS());	// X-A -> Y

			if(fdTemp2.IsMemberOf(fdsetTemp1))
			{
				fdsetTemp1.Remove(fdTemp1);
				fdsetTemp1+=fdTemp2;
//				fdsetTemp1.AddFD(fdTemp2);

			}
		}
	}

	m_nIndex=nBackupIndex;
	m_rPosition=rBackupPosition;
	return fdsetTemp1;
}

CFDSet CFDSet::CanonicalCover()
{
	CString strTemp;

	strTemp=this->GetstrFD();

	CFDSet fdsetG;
	CFD fdTrue;

	CAttributeSet attsetX, attsetY;
	CAttribute attA;

	CFD fdFd;

	const int nSizeFDSet=m_fdList.GetCount();
	int nSizeAttSet;
	this->GoHeadPosition();
	for(int i=0; i<nSizeFDSet; i++)
	{
		fdFd=this->GetFD();
		this->operator++();

		strTemp=fdFd.GetstrFD();

		attsetX=fdFd.GetLHS();
		attsetY=fdFd.GetRHS();
		
		nSizeAttSet=attsetY.GetCount();
		attsetY.GoHeadPosition();
		for(int j=0; j<nSizeAttSet; j++)
		{
			attA=attsetY.GetAtt();
			attsetY++;
			
			fdTrue.SetLHS(attsetX);
			fdTrue.SetRHS(attA);

			strTemp=fdTrue.GetstrFD();

			fdsetG.AddFD(fdTrue);
		}

		strTemp=fdsetG.GetstrFD();

	}			

	return fdsetG;
}

CFDSet CFDSet::MinumumCover()
{
	const int nBackupIndex=m_nIndex;
	const POSITION rBackupPosition=m_rPosition;

	CString strTemp;

	strTemp=this->GetstrFD();

	CFDSet fdsetG;

	CFD fdX, fdXPlus;
	int nLenOfFDList=m_fdList.GetCount();


	//step 1
	this->GoHeadPosition();
	for(int i=0; i<nLenOfFDList; i++)
	{
		fdX=this->GetFD();
		this->operator ++();

		strTemp=fdX.GetstrFD();
		
		fdXPlus.SetLHS(fdX.GetLHS());
		fdXPlus.SetRHS(fdX.GetLHS().Closure(*this));

		strTemp=fdXPlus.GetstrFD();

		fdsetG.AddFD(fdXPlus);
	}

	strTemp=fdsetG.GetstrFD();

	//step2
	nLenOfFDList=fdsetG.GetCount();
	fdsetG.GoHeadPosition();
	for(i=0; i<nLenOfFDList; i++)
	{
		fdX=fdsetG.GetFD();
		fdsetG++;

		strTemp=fdX.GetstrFD();

		if(fdX.IsRedundant(fdsetG)==TRUE)
		{
			fdsetG.Remove(fdX);
			i--;
			nLenOfFDList--;
		}
	}

	CFDSet fdsetM;

	int nNumFD=fdsetG.GetCount();
	fdsetG.GoHeadPosition();
	for(int cJ=0; cJ<nNumFD; cJ++)
	{
		CFD fdFD=fdsetG.GetFD();
		fdsetG++;

		fdFD.SetRHS(fdFD.GetRHS()-fdFD.GetLHS());

		fdsetM.AddFD(fdFD);
	}

	strTemp=fdsetM.GetstrFD();

	m_nIndex=nBackupIndex;
	m_rPosition=rBackupPosition;
	return fdsetM;
}

BOOL CFDSet::HasMember(CFD fdInput)
{
	if(fdInput.GetLHS().Closure(*this).IsMemberOf(fdInput.GetRHS()))
		return TRUE;
	return FALSE;
}

BOOL CFDSet::IsDerives(CFDSet fdsetInput)
{
	const int nBackupIndex=m_nIndex;
	const POSITION rBackupPosition=m_rPosition;
	
	const int nLenOfFDList=fdsetInput.GetCount();
	CFD fdTemp;

	fdsetInput.GoHeadPosition();
	for(int i=0; i<nLenOfFDList; i++)
	{
		fdTemp=fdsetInput.GetFD();
		fdsetInput++;

		if(fdTemp.IsMemberOf(*this)==FALSE)
			return FALSE;
	}

	m_nIndex=nBackupIndex;
	m_rPosition=rBackupPosition;
	return TRUE;
}

BOOL CFDSet::IsDirectDerives(CAttributeSet attsetInput1,CAttributeSet attsetInput2)
{
	const int nBackupIndex=m_nIndex;
	const POSITION rBackupPosition=m_rPosition;

	CFDSet fdsetTemp=this->NonRedundantCover();
	fdsetTemp-=this->EquivClass(attsetInput1);
    
	CFD fdTemp;
	fdTemp.SetLHS(attsetInput1);
	fdTemp.SetRHS(attsetInput2);

	if(fdTemp.IsMemberOf(fdsetTemp))
	{
		m_nIndex=nBackupIndex;
		m_rPosition=rBackupPosition;
		
		return TRUE;
	}

	m_nIndex=nBackupIndex;
	m_rPosition=rBackupPosition;
	return FALSE;
}

BOOL CFDSet::IsEquivalent(CFDSet fdsetInput)
{
	if(IsDerives(*this)==TRUE)
		if(IsDerives(fdsetInput)==TRUE)
			return TRUE;
	return FALSE;
}

BOOL CFDSet::IsEquivalent(CAttributeSet attsetInput1, CAttributeSet attsetInput2)
{
	CFD fdTemp1;
	CFD fdTemp2;

	fdTemp1.SetLHS(attsetInput1);
	fdTemp1.SetRHS(attsetInput2);

	fdTemp2.SetLHS(attsetInput2);
	fdTemp2.SetRHS(attsetInput1);

	if(this->HasMember(fdTemp1)==TRUE)
		if(this->HasMember(fdTemp2)==TRUE)
			return TRUE;
	return FALSE;
}

BOOL CFDSet::IsRedundant()
{
	const int nBackupIndex=m_nIndex;
	const POSITION rBackupPosition=m_rPosition;

	CFD fdTemp;
	CFDSet fdsetTemp=*this;
	int nLenOfFDList=m_fdList.GetCount();

	this->GoHeadPosition();
	for(int i=0; i<nLenOfFDList; i++)
	{
		fdTemp=this->GetFD();
		this->operator++();

		fdsetTemp.Remove(fdTemp);
		if(fdTemp.IsMemberOf(fdsetTemp)==TRUE)
			return TRUE;
	}

	m_nIndex=nBackupIndex;
	m_rPosition=rBackupPosition;

	return FALSE;
}

CFDSet CFDSet::EquivClass(CAttributeSet attsetInput)
{
	const int nBackupIndex=m_nIndex;
	const POSITION rBackupPosition=m_rPosition;

	CFDSet fdsetTemp;
	CFD fdTemp;
	int nLenOfFDList=m_fdList.GetCount();

	this->GoHeadPosition();
	for(int i=0; i<nLenOfFDList; i++)
	{
		fdTemp=this->GetFD();
		this->operator++();

		if(this->IsEquivalent(fdTemp.GetLHS(), attsetInput))
			fdsetTemp+=fdTemp;
	}

	m_nIndex=nBackupIndex;
	m_rPosition=rBackupPosition;
	return fdsetTemp;
}

CFDSet CFDSet::operator +(CFDSet& fdsetInput)
{
	const int nBackupIndex=fdsetInput.GetIndex();

	CFD fdTemp;
	CFDSet fdsetTemp;;

	fdsetTemp=*this;

	fdsetInput.GoHeadPosition();
	for(int i=0; i<fdsetInput.GetCount(); i++)
	{
		fdTemp=fdsetInput.GetFD();
		fdsetTemp.AddFD(fdTemp);
		fdsetInput++;
	}

	m_nIndex=0;
	m_rPosition=m_fdList.GetHeadPosition();

	fdsetInput.GoIndex(nBackupIndex);
	return fdsetTemp;
}

CFDSet CFDSet::operator-(CFD& fdInput)
{
	CFDSet fdsetTemp=*this;
	fdsetTemp.Remove(fdInput);

	return fdsetTemp;
}

CFDSet CFDSet::operator-(CFDSet& fdsetInput)
{
	const int nBackupIndex=fdsetInput.GetIndex();

	CFD fdTemp;
	CFDSet fdsetTemp=*this;
	const int nLenOfFDList=fdsetInput.GetCount();

	fdsetInput.GoHeadPosition();
	for(int i=0; i<nLenOfFDList; i++)
	{
		fdTemp=fdsetInput.GetFD();
		fdsetInput++;
		
		fdsetTemp.Remove(fdTemp);
	}

	fdsetInput.GoIndex(nBackupIndex);
	return fdsetTemp;
}

CFDSet& CFDSet::operator=(CFD& fdInput)
{
	*this=CFDSet(fdInput);

	return *this;		
}

CFDSet& CFDSet::operator=(CFDSet& fdsetInput)
{
	const int nBackupIndex=fdsetInput.GetIndex();
	if(this==&fdsetInput)
		return *this;

	const int nInLength=fdsetInput.GetCount();
	CFD fdTemp;
	m_fdList.RemoveAll();

	fdsetInput.GoHeadPosition();
	for(int i=0; i<nInLength; i++)
	{
		fdTemp=fdsetInput.GetFD();
		m_fdList.AddTail(fdTemp);
		fdsetInput++;
	}

	m_nIndex=0;
	m_rPosition=m_fdList.GetHeadPosition();

	fdsetInput.GoIndex(nBackupIndex);
	return *this;		
}

CFDSet& CFDSet::operator +=(CFD& fdInput)
{
	*this=*this+CFDSet(fdInput);

	return *this;
}

CFDSet& CFDSet::operator +=(CFDSet& fdsetInput)
{
	const int nBackupIndex=fdsetInput.GetIndex();

	*this=*this+fdsetInput;

	fdsetInput.GoIndex(nBackupIndex);
	return *this;
}

CFDSet& CFDSet::operator -=(CFD& fdInput)
{
	this->Remove(fdInput);

	return *this;
}

CFDSet& CFDSet::operator -=(CFDSet& fdsetInput)
{
	const int nBackupIndex=fdsetInput.GetIndex();

	*this=*this-fdsetInput;

	fdsetInput.GoIndex(nBackupIndex);
	return *this;
}

void CFDSet::operator ++()
{
	m_nIndex++;
	if(m_nIndex>=m_fdList.GetCount())
		m_nIndex=m_fdList.GetCount()-1;
	else
		m_rPosition=m_fdList.FindIndex(m_nIndex);
}

void CFDSet::operator --()
{
	m_nIndex--;
	if(m_nIndex<0)
		m_nIndex=0;
	else
		m_rPosition=m_fdList.FindIndex(m_nIndex);
}


void CFDSet::operator ++(int)
{
	operator++();
}

void CFDSet::operator --(int)
{
	operator--();
}

BOOL CFDSet::operator ==(CFDSet& fdsetInput)
{
	const int nBackupIndex=fdsetInput.GetIndex();

	if(m_fdList.GetCount()!=fdsetInput.GetCount())
		return FALSE;

	int nLenOfFDList1=fdsetInput.GetCount();
	int nLenOfFDList2=m_fdList.GetCount();
	CFD fdTemp1;
	CFD fdTemp2;
	BOOL bSam=FALSE;
	
	fdsetInput.GoHeadPosition();
	for(int i=0; i<nLenOfFDList1; i++)
	{
		fdTemp1=fdsetInput.GetFD();
		fdsetInput++;

		bSam=FALSE;

		this->GoHeadPosition();
		for(int j=0; j<nLenOfFDList2; j++)
		{
			fdTemp2=this->GetFD();
			this->operator ++();

/*			if(fdTemp1!=fdTemp2)
			{
				fdsetInput.GoIndex(nBackupIndex);
				return FALSE;
			}*/

			if(fdTemp1==fdTemp2)
			{
				bSam=TRUE;
				break;
			}
		}
		if(bSam==FALSE)
		{
			fdsetInput.GoIndex(nBackupIndex);
			return FALSE;
		}

	}

	fdsetInput.GoIndex(nBackupIndex);
	return TRUE;
}

BOOL CFDSet::operator !=(CFDSet& fdsetInput)
{
	const int nBackupIndex=fdsetInput.GetIndex();

	return !(this->operator==(fdsetInput));
	
	fdsetInput.GoIndex(nBackupIndex);
}

CString CFDSet::GetstrFD()
{
	CString strTemp;
	const int nListLength=m_fdList.GetCount();
	
	GoHeadPosition();
	for(int i=0; i<nListLength; i++)
	{
		strTemp+=GetFD().GetstrFD()+"\r\n";
		operator++();
	}

	return strTemp;
}